# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QGroupBox, QHBoxLayout, QSlider, QLabel, QCheckBox, \
    QVBoxLayout

from lisp.backend.audio_utils import db_to_linear, linear_to_db
from lisp.modules.gst_backend.elements.volume import Volume
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.widgets import QMuteButton
from lisp.ui.ui_utils import translate


class VolumeSettings(SettingsPage):
    ELEMENT = Volume
    Name = ELEMENT.Name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.normal = 1

        self.volumeBox = QGroupBox(self)
        self.volumeBox.setLayout(QHBoxLayout())
        self.layout().addWidget(self.volumeBox)

        self.muteButton = QMuteButton(self.volumeBox)
        self.volumeBox.layout().addWidget(self.muteButton)

        self.volume = QSlider(self.volumeBox)
        self.volume.setRange(-1000, 100)
        self.volume.setPageStep(1)
        self.volume.setOrientation(Qt.Horizontal)
        self.volume.valueChanged.connect(self.volume_changed)
        self.volumeBox.layout().addWidget(self.volume)

        self.volumeLabel = QLabel(self.volumeBox)
        self.volumeLabel.setAlignment(Qt.AlignCenter)
        self.volumeBox.layout().addWidget(self.volumeLabel)

        self.volumeBox.layout().setStretch(0, 1)
        self.volumeBox.layout().setStretch(1, 4)
        self.volumeBox.layout().setStretch(2, 1)

        self.normalBox = QGroupBox(self)
        self.normalBox.setLayout(QHBoxLayout())
        self.layout().addWidget(self.normalBox)

        self.normalLabel = QLabel(self.normalBox)
        self.normalLabel.setAlignment(Qt.AlignCenter)
        self.normalBox.layout().addWidget(self.normalLabel)

        self.normalReset = QCheckBox(self.normalBox)
        self.normalBox.layout().addWidget(self.normalReset)
        self.normalBox.layout().setAlignment(self.normalReset, Qt.AlignCenter)

        self.retranslateUi()

    def retranslateUi(self):
        self.volumeBox.setTitle(translate('VolumeSettings', 'Volume'))
        self.volumeLabel.setText('0.0 dB')
        self.normalBox.setTitle(
            translate('VolumeSettings', 'Normalized volume'))
        self.normalLabel.setText('0.0 dB')
        self.normalReset.setText(translate('VolumeSettings', 'Reset'))

    def enable_check(self, enable):
        for box in [self.normalBox, self.volumeBox]:
            box.setCheckable(enable)
            box.setChecked(False)

    def get_settings(self):
        settings = {}
        checkable = self.volumeBox.isCheckable()

        if not (checkable and not self.volumeBox.isChecked()):
            settings['volume'] = db_to_linear(self.volume.value() / 10)
            settings['mute'] = self.muteButton.isMute()
        if not (checkable and not self.normalBox.isChecked()):
            if self.normalReset.isChecked():
                settings['normal_volume'] = 1
                # If the apply button is pressed, show the correct value
                self.normalLabel.setText('0 dB')
            else:
                settings['normal_volume'] = self.normal

        return settings

    def load_settings(self, settings):
        self.volume.setValue(linear_to_db(settings.get('volume', 1)) * 10)
        self.muteButton.setMute(settings.get('mute', False))
        self.normal = settings.get('normal_volume', 1)
        self.normalLabel.setText(
            str(round(linear_to_db(self.normal), 3)) + ' dB')

    def volume_changed(self, value):
        self.volumeLabel.setText(str(value / 10.0) + ' dB')



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy

from PyQt5 import QtCore
from PyQt5.QtWidgets import QDialog, QTabWidget, QDialogButtonBox

from lisp.core.class_based_registry import ClassBasedRegistry
from lisp.core.singleton import Singleton
from lisp.core.util import deep_update
from lisp.cues.cue import Cue
from lisp.ui.settings.settings_page import SettingsPage, CueSettingsPage
from lisp.ui.ui_utils import translate


class CueSettingsRegistry(ClassBasedRegistry, metaclass=Singleton):
    def add_item(self, item, ref_class=Cue):
        if not issubclass(item, SettingsPage):
            raise TypeError('item must be a SettingPage subclass, '
                            'not {0}'.format(item.__name__))
        if not issubclass(ref_class, Cue):
            raise TypeError('ref_class must be Cue or a subclass, not {0}'
                            .format(ref_class.__name__))

        return super().add_item(item, ref_class)

    def filter(self, ref_class=Cue):
        return super().filter(ref_class)

    def clear_class(self, ref_class=Cue):
        return super().filter(ref_class)


class CueSettings(QDialog):
    on_apply = QtCore.pyqtSignal(dict)

    def __init__(self, cue=None, cue_class=None, **kwargs):
        """

        :param cue: Target cue, or None for multi-editing
        :param cue_class: when cue is None, used to specify the reference class
        """
        super().__init__(**kwargs)

        if cue is not None:
            cue_class = cue.__class__
            cue_properties = deepcopy(cue.properties())
            self.setWindowTitle(cue_properties['name'])
        else:
            cue_properties = {}
            if cue_class is None:
                cue_class = Cue

        self.setWindowModality(QtCore.Qt.ApplicationModal)
        self.setMaximumSize(635, 530)
        self.setMinimumSize(635, 530)
        self.resize(635, 530)

        self.sections = QTabWidget(self)
        self.sections.setGeometry(QtCore.QRect(5, 10, 625, 470))

        def sk(widget):
            # Sort-Key function
            return translate('SettingsPageName', widget.Name)

        for widget in sorted(CueSettingsRegistry().filter(cue_class), key=sk):
            if issubclass(widget, CueSettingsPage):
                settings_widget = widget(cue_class)
            else:
                settings_widget = widget()

            settings_widget.load_settings(cue_properties)
            settings_widget.enable_check(cue is None)
            self.sections.addTab(settings_widget,
                                 translate('SettingsPageName',
                                           settings_widget.Name))

            self.dialogButtons = QDialogButtonBox(self)
            self.dialogButtons.setGeometry(10, 490, 615, 30)
            self.dialogButtons.setStandardButtons(QDialogButtonBox.Cancel |
                                                  QDialogButtonBox.Ok |
                                                  QDialogButtonBox.Apply)

            self.dialogButtons.rejected.connect(self.reject)
            self.dialogButtons.accepted.connect(self.accept)
            apply = self.dialogButtons.button(QDialogButtonBox.Apply)
            apply.clicked.connect(self.apply)

    def load_settings(self, settings):
        for n in range(self.sections.count()):
            self.sections.widget(n).load_settings(settings)

    def get_settings(self):
        settings = {}

        for n in range(self.sections.count()):
            deep_update(settings, self.sections.widget(n).get_settings())

        return settings

    def apply(self):
        self.on_apply.emit(self.get_settings())

    def accept(self):
        self.apply()
        super().accept()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QGroupBox, QHBoxLayout, QSlider, QLabel, QVBoxLayout

from lisp.modules.gst_backend.elements.audio_pan import AudioPan
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class AudioPanSettings(SettingsPage):

    ELEMENT = AudioPan
    Name = ELEMENT.Name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.panBox = QGroupBox(self)
        self.panBox.setGeometry(0, 0, self.width(), 80)
        self.panBox.setLayout(QHBoxLayout(self.panBox))
        self.layout().addWidget(self.panBox)

        self.panSlider = QSlider(self.panBox)
        self.panSlider.setRange(-10, 10)
        self.panSlider.setPageStep(1)
        self.panSlider.setOrientation(Qt.Horizontal)
        self.panSlider.valueChanged.connect(self.pan_changed)
        self.panBox.layout().addWidget(self.panSlider)

        self.panLabel = QLabel(self.panBox)
        self.panLabel.setAlignment(Qt.AlignCenter)
        self.panBox.layout().addWidget(self.panLabel)

        self.panBox.layout().setStretch(0, 5)
        self.panBox.layout().setStretch(1, 1)

        self.retransaleUi()

    def retransaleUi(self):
        self.panBox.setTitle(translate('AudioPanSettings', 'Audio Pan'))
        self.panLabel.setText(translate('AudioPanSettings', 'Center'))

    def enable_check(self, enable):
        self.panBox.setCheckable(enable)
        self.panBox.setChecked(False)

    def get_settings(self):
        if not (self.panBox.isCheckable() and not self.panBox.isChecked()):
            return {'pan': self.panSlider.value() / 10}

        return {}

    def load_settings(self, settings):
        self.panSlider.setValue(settings.get('pan', 0.5) * 10)

    def pan_changed(self, value):
        if value < 0:
            position = translate('AudioPanSettings', 'Left')
        elif value > 0:
            position = translate('AudioPanSettings', 'Right')
        else:
            position = translate('AudioPanSettings', 'Center')

        self.panLabel.setText('{0} - {1}'.format(value, position))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt, QT_TRANSLATE_NOOP
from PyQt5.QtWidgets import QComboBox
from PyQt5.QtWidgets import QVBoxLayout, QGroupBox

from lisp.application import Application
from lisp.core.has_properties import Property
from lisp.cues.cue import Cue, CueAction
from lisp.ui.settings.cue_settings import CueSettingsRegistry
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class StopAll(Cue):
    Name = QT_TRANSLATE_NOOP('CueName', 'Stop-All')

    action = Property(default=CueAction.Stop.value)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = translate('CueName', self.Name)

    def __start__(self, fade=False):
        for cue in Application().cue_model:
            action = self.__adjust_action(cue, CueAction(self.action))
            if action:
                cue.execute(action=action)

        return False

    def __adjust_action(self, cue, action, fade=False):
        if action in cue.CueActions:
            return action
        elif action is CueAction.FadeOutPause:
            return self.__adjust_action(cue, CueAction.Pause, True)
        elif action is CueAction.Pause and fade:
            return self.__adjust_action(cue, CueAction.FadeOutStop)
        elif action is CueAction.FadeOutInterrupt:
            return self.__adjust_action(cue, CueAction.Interrupt)
        elif action is CueAction.FadeOutStop:
            return self.__adjust_action(cue, CueAction.Stop)

        return None


class StopAllSettings(SettingsPage):
    Name = QT_TRANSLATE_NOOP('SettingsPageName', 'Stop Settings')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.group = QGroupBox(self)
        self.group.setLayout(QVBoxLayout(self.group))
        self.layout().addWidget(self.group)

        self.actionCombo = QComboBox(self.group)
        for action in [CueAction.Stop, CueAction.FadeOutStop, CueAction.Pause,
                       CueAction.FadeOutPause, CueAction.Interrupt,
                       CueAction.FadeOutInterrupt]:
            self.actionCombo.addItem(
                translate('CueAction', action.name), action.value)
        self.group.layout().addWidget(self.actionCombo)

        self.retranslateUi()

    def retranslateUi(self):
        self.group.setTitle(translate('StopAll', 'Stop Action'))

    def enable_check(self, enabled):
        self.group.setCheckable(enabled)
        self.group.setChecked(False)

    def get_settings(self):
        conf = {}

        if not (self.group.isCheckable() and not self.group.isChecked()):
            conf['action'] = self.actionCombo.currentData()

        return conf

    def load_settings(self, settings):
        self.actionCombo.setCurrentText(
            translate('CueAction', settings.get('action', '')))


CueSettingsRegistry().add_item(StopAllSettings, StopAll)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import socket

from lisp.core.configuration import config
from lisp.core.module import Module
from lisp.modules.remote.controller import RemoteController


class Remote(Module):
    """Provide remote control over cues via RPCs calls."""

    def __init__(self):
        ip = config['Remote']['BindIp']
        port = int(config['Remote']['BindPort'])

        RemoteController(ip=ip, port=port)
        RemoteController().start()

    def terminate(self):
        RemoteController().stop()


def compose_uri(url, port, directory='/'):
    return 'http://' + url + ':' + str(port) + directory


# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtWidgets import QMessageBox


class QDetailedMessageBox(QMessageBox):
    """QMessageBox providing statics methods to show detailed-text"""

    @staticmethod
    def dcritical(title, text, detailed_text, parent=None):
        """MessageBox with "Critical" icon"""
        QDetailedMessageBox.dgeneric(title,
                                     text,
                                     detailed_text,
                                     QMessageBox.Critical,
                                     parent)

    @staticmethod
    def dwarning(title, text, detailed_text, parent=None):
        """MessageBox with "Warning" icon"""
        QDetailedMessageBox.dgeneric(title,
                                     text,
                                     detailed_text,
                                     QMessageBox.Warning,
                                     parent)

    @staticmethod
    def dinformation(title, text, detailed_text, parent=None):
        """MessageBox with "Information" icon"""
        QDetailedMessageBox.dgeneric(title,
                                     text,
                                     detailed_text,
                                     QMessageBox.Information,
                                     parent)

    @staticmethod
    def dgeneric(title, text, detail_text, icon, parent=None):
        """Build and show a MessageBox with detailed text"""
        messageBox = QMessageBox(parent)

        messageBox.setIcon(icon)
        messageBox.setWindowTitle(title)
        messageBox.setText(text)
        # Show the detail text only if not an empty string
        if detail_text.strip() != '':
            messageBox.setDetailedText(detail_text)
        messageBox.addButton(QMessageBox.Ok)
        messageBox.setDefaultButton(QMessageBox.Ok)

        messageBox.exec_()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import re
import socket
from collections import Mapping
from enum import Enum
from os import listdir
from os.path import isdir, exists, join

import functools


def deep_update(d1, d2):
    """Recursively update d1 with d2"""
    for key in d2:
        if key not in d1:
            d1[key] = d2[key]
        elif isinstance(d2[key], Mapping) and isinstance(d1[key], Mapping):
            d1[key] = deep_update(d1[key], d2[key])
        else:
            d1[key] = d2[key]

    return d1


def find_packages(path='.'):
    """List the python packages in the given directory."""

    return [d for d in listdir(path) if isdir(join(path, d)) and
            exists(join(path, d, '__init__.py'))]


def time_tuple(millis):
    """Split the given time in a tuple.

    :param millis: Number of milliseconds
    :type millis: int

    :return (hours, minutes, seconds, milliseconds)
    """
    seconds, millis = divmod(millis, 1000)
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)

    return hours, minutes, seconds, millis


def strtime(time, accurate=False):
    """Return a string from the given milliseconds time.

    :returns:
        hh:mm:ss when > 59min
        mm:ss:00 when < 1h and accurate=False
        mm:ss:z0 when < 1h and accurate=True
    """
    # Cast time to int to avoid formatting problems
    time = time_tuple(int(time))
    if time[0] > 0:
        return '{:02}:{:02}:{:02}'.format(*time[:-1])
    elif accurate:
        return '{:02}:{:02}.{}0'.format(time[1], time[2], time[3] // 100)
    else:
        return '{:02}:{:02}.00'.format(*time[1:3])


def compose_http_url(url, port, directory='/'):
    """Compose an http URL."""
    return 'http://' + url + ':' + str(port) + directory


def greatest_common_superclass(instances):
    classes = [type(x).mro() for x in instances]
    for x in classes[0]:
        if all(x in mro for mro in classes):
            return x


def get_lan_ip():
    """Return active interface LAN IP, or localhost if no address is assigned.

    From: http://stackoverflow.com/a/28950776/5773767
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # Doesn't have to be reachable
        s.connect(('10.255.255.255', 0))
        ip = s.getsockname()[0]
    except:
        ip = '127.0.0.1'
    finally:
        s.close()
    return ip


def subclasses(cls):
    for subclass in cls.__subclasses__():
        yield from subclasses(subclass)
        yield subclass


def weak_call_proxy(weakref):
    def proxy(*args, **kwargs):
        if weakref() is not None:
            weakref()(*args, **kwargs)

    return proxy


def natural_keys(text):
    """Turn a string into a list of string and number chunks.

    "z23a" -> ["z", 23, "a"]

    From: http://stackoverflow.com/a/5967539/5773767

    .. highlight::

        l = ['something1', 'something17', 'something4']
        l.sort(key=natural_keys) # sorts in human order
        ['something1', 'something4', 'something17']
    """
    return [int(c) if c.isdigit() else c for c in re.split('(\d+)', text)]


def rhasattr(obj, attr):
    """Check object's attribute, can use dot notation.

    .. highlight::

        class A: pass
        a = A()
        a.b = A()
        a.b.c = 42
        hasattr(a, 'b.c')  # True
    """
    return functools.reduce(hasattr, attr.split('.'), obj)


def rsetattr(obj, attr, value):
    """Set object's attribute, can use dot notation.

    From: http://stackoverflow.com/a/31174427/5773767

    .. highlight::

        class A: pass
        a = A()
        a.b = A()
        a.b.c = 0
        rsetattr(a, 'b.c', 42)
        a.b.c  # 42
    """
    pre, _, post = attr.rpartition('.')
    setattr(rgetattr(obj, pre) if pre else obj, post, value)


rgetattr_sentinel = object()


def rgetattr(obj, attr, default=rgetattr_sentinel):
    """Get object's attribute, can use dot notation.

    From: http://stackoverflow.com/a/31174427/5773767

    .. highlight::

        class A: pass
        a = A()
        a.b = A()
        a.b.c = 24
        rgetattr(a, 'b.c')  # 42
    """
    if default is rgetattr_sentinel:
        _getattr = getattr
    else:
        def _getattr(obj, name):
            return getattr(obj, name, default)

    return functools.reduce(_getattr, attr.split('.'), obj)


class EqEnum(Enum):
    """Value-comparable Enum.

    EqEnum members can be compared for equality with their values:

    .. highlight::

        class E(EqEnum):
            A = 10

        class E2(EqEnum):
            A2 = 10

        # Equality NOT identity
        E.A == 10  # True
        E.A is 10  # False

        E.A == E.A2  # False
    """

    def __eq__(self, other):
        if not isinstance(other, Enum):
            return self.value == other

        return super().__eq__(other)


class FunctionProxy:
    """Allow to mask a function as an Object.

    Can be useful in enum.Enum (Python enumeration) to have callable values.
    """
    def __init__(self, function):
        self.function = function

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt, QEvent
from PyQt5.QtWidgets import QStyledItemDelegate, QComboBox, QSpinBox, \
    QLineEdit, QStyle, QDialog

from lisp.application import Application
from lisp.cues.cue import CueAction
from lisp.ui.qmodels import CueClassRole
from lisp.ui.ui_utils import translate
from lisp.ui.widgets import CueActionComboBox


class LabelDelegate(QStyledItemDelegate):
    def _text(self, painter, option, index):
        return ''

    def paint(self, painter, option, index):
        # Add 4px of left an right padding
        option.rect.adjust(4, 0, 4, 0)

        text = self._text(painter, option, index)
        text = option.fontMetrics.elidedText(text, Qt.ElideRight,
                                             option.rect.width())

        painter.save()

        if option.state & QStyle.State_Selected:
            painter.setBrush(option.palette.highlight())

            pen = painter.pen()
            pen.setBrush(option.palette.highlightedText())
            painter.setPen(pen)

        painter.drawText(option.rect, option.displayAlignment, text)

        painter.restore()


class ComboBoxDelegate(LabelDelegate):
    def __init__(self, options=(), tr_context=None, **kwargs):
        super().__init__(**kwargs)
        self.options = options
        self.tr_context = tr_context

    def _text(self, painter, option, index):
        return translate(self.tr_context, index.data())

    def paint(self, painter, option, index):
        option.displayAlignment = Qt.AlignHCenter | Qt.AlignVCenter
        super().paint(painter, option, index)

    def createEditor(self, parent, option, index):
        editor = QComboBox(parent)
        editor.setFrame(False)
        for option in self.options:
            editor.addItem(translate(self.tr_context, option), option)

        return editor

    def setEditorData(self, comboBox, index):
        value = index.model().data(index, Qt.EditRole)
        comboBox.setCurrentText(translate(self.tr_context, value))

    def setModelData(self, comboBox, model, index):
        model.setData(index, comboBox.currentData(), Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class SpinBoxDelegate(QStyledItemDelegate):
    def __init__(self, minimum=-100, maximum=100, step=1, **kwargs):
        super().__init__(**kwargs)

        self.minimum = minimum
        self.maximum = maximum
        self.step = step

    def createEditor(self, parent, option, index):
        editor = QSpinBox(parent)
        editor.setRange(self.minimum, self.maximum)
        editor.setSingleStep(self.step)

        return editor

    def setEditorData(self, spinBox, index):
        value = index.model().data(index, Qt.EditRole)
        if isinstance(value, int):
            spinBox.setValue(value)

    def setModelData(self, spinBox, model, index):
        spinBox.interpretText()
        model.setData(index, spinBox.value(), Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class LineEditDelegate(QStyledItemDelegate):
    def __init__(self, max_length=-1, validator=None, **kwargs):
        super().__init__(**kwargs)

        self.max_length = max_length
        self.validator = validator

    def createEditor(self, parent, option, index):
        editor = QLineEdit(parent)
        if self.max_length > 0:
            editor.setMaxLength(self.max_length)
        if self.validator is not None:
            editor.setValidator(self.validator)
        editor.setFrame(False)

        return editor

    def setEditorData(self, lineEdit, index):
        value = index.model().data(index, Qt.EditRole)
        lineEdit.setText(str(value))

    def setModelData(self, lineEdit, model, index):
        model.setData(index, lineEdit.text(), Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class CueActionDelegate(LabelDelegate):
    Mode = CueActionComboBox.Mode

    def __init__(self, cue_class=None, mode=Mode.Action, **kwargs):
        super().__init__(**kwargs)
        self.cue_class = cue_class
        self.mode = mode

    def _text(self, painter, option, index):
        value = index.data(Qt.EditRole)
        if self.mode == CueActionDelegate.Mode.Action:
            name = value.name
        elif self.mode == CueActionDelegate.Mode.Name:
            name = value
        else:
            name = CueAction(value).name

        return translate('CueAction', name)

    def paint(self, painter, option, index):
        option.displayAlignment = Qt.AlignHCenter | Qt.AlignVCenter
        super().paint(painter, option, index)

    def createEditor(self, parent, option, index):
        if self.cue_class is None:
            self.cue_class = index.data(CueClassRole)

        editor = CueActionComboBox(self.cue_class,
                                   mode=self.mode,
                                   parent=parent)
        editor.setFrame(False)

        return editor

    def setEditorData(self, comboBox, index):
        value = index.data(Qt.EditRole)
        if self.mode == CueActionDelegate.Mode.Action:
            action = value
        elif self.mode == CueActionDelegate.Mode.Name:
            action = CueAction[value]
        else:
            action = CueAction(value)

        comboBox.setCurrentText(translate('CueAction', action.name))

    def setModelData(self, comboBox, model, index):
        model.setData(index, comboBox.currentData(), Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class CueSelectionDelegate(LabelDelegate):
    def __init__(self, cue_select_dialog, **kwargs):
        super().__init__(**kwargs)
        self.cue_select = cue_select_dialog

    def _text(self, painter, option, index):
        cue = Application().cue_model.get(index.data())
        if cue is not None:
            return '{} | {}'.format(cue.index, cue.name)

        return 'UNDEF'

    def editorEvent(self, event, model, option, index):
        if event.type() == QEvent.MouseButtonDblClick:
            if self.cue_select.exec_() == QDialog.Accepted:
                cue = self.cue_select.selected_cue()
                if cue is not None:
                    model.setData(index, cue.id, Qt.EditRole)
                    model.setData(index, cue.__class__, CueClassRole)
            return True

        return super().editorEvent(event, model, option, index)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QComboBox
from PyQt5.QtWidgets import QDialog, QInputDialog, QMessageBox, QListWidget, \
    QPushButton, QVBoxLayout, QGridLayout, QDialogButtonBox, QWidget, QLabel, \
    QLineEdit
from PyQt5.QtWidgets import QFileDialog
from PyQt5.QtWidgets import QHBoxLayout

from lisp.application import Application
from lisp.core.util import natural_keys
from lisp.cues.cue import Cue
from lisp.cues.cue_factory import CueFactory
from lisp.modules.presets.lib import preset_exists, export_presets, \
    import_presets, import_has_conflicts, PresetExportError,\
    PresetImportError, scan_presets, delete_preset, write_preset,\
    rename_preset, load_preset, load_on_cues
from lisp.ui.mainwindow import MainWindow
from lisp.ui.settings.cue_settings import CueSettings, CueSettingsRegistry
from lisp.ui.ui_utils import translate
from lisp.ui.widgets import QDetailedMessageBox


def preset_error(exception, text, parent=None):
    QDetailedMessageBox.dcritical(
        translate('Presets', 'Presets'),
        text,
        str(exception),
        parent=parent,
    )


def scan_presets_error(exception, parent=None):
    preset_error(
        exception,
        translate('Presets', 'Cannot scan presets'),
        parent=parent
    )


def delete_preset_error(exception, name, parent=None):
    preset_error(
        exception,
        translate('Presets', 'Error while deleting preset "{}"').format(name),
        parent=parent
    )


def load_preset_error(exception, name, parent=None):
    preset_error(
        exception,
        translate('Presets', 'Cannot load preset "{}"').format(name),
        parent=parent
    )


def write_preset_error(exception, name, parent=None):
    preset_error(
        exception,
        translate('Presets', 'Cannot save preset "{}"').format(name),
        parent=parent
    )


def rename_preset_error(exception, name, parent=None):
    preset_error(
        exception,
        translate('Presets', 'Cannot rename preset "{}"').format(name),
        parent=parent
    )


def select_preset_dialog():
    try:
        presets = tuple(sorted(scan_presets(), key=natural_keys))

        if presets:
            item, confirm = QInputDialog.getItem(
                MainWindow(),
                translate('Presets', 'Select Preset'), '',
                presets
            )

            if confirm:
                return item
    except OSError as e:
        scan_presets_error(e, parent=MainWindow())


def check_override_dialog(preset_name):
    answer = QMessageBox.question(
        MainWindow(),
        translate('Presets', 'Presets'),
        translate('Presets', 'Preset already exists, overwrite?'),
        buttons=QMessageBox.Yes | QMessageBox.Cancel
    )

    return answer == QMessageBox.Yes


def save_preset_dialog(base_name=''):
    name, confirm = QInputDialog.getText(
        MainWindow(),
        translate('Presets', 'Presets'),
        translate('Presets', 'Preset name'),
        text=base_name
    )

    if confirm:
        return name


class PresetsDialog(QDialog):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.resize(500, 400)
        self.setMaximumSize(self.size())
        self.setMinimumSize(self.size())
        self.setModal(True)
        self.setLayout(QGridLayout())

        # TODO: natural sorting (QStringListModel + QListView + ProxyModel)
        self.presetsList = QListWidget(self)
        self.presetsList.setAlternatingRowColors(True)
        self.presetsList.setFocusPolicy(Qt.NoFocus)
        self.presetsList.setSortingEnabled(True)
        self.presetsList.setSelectionMode(QListWidget.ExtendedSelection)
        self.presetsList.itemSelectionChanged.connect(self.__selection_changed)
        self.presetsList.itemDoubleClicked.connect(self.__edit_preset)
        self.layout().addWidget(self.presetsList, 0, 0)

        # Preset buttons
        self.presetsButtons = QWidget(self)
        self.presetsButtons.setLayout(QVBoxLayout())
        self.presetsButtons.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self.presetsButtons, 0, 1)
        self.layout().setAlignment(self.presetsButtons, Qt.AlignTop)

        self.addPresetButton = QPushButton(self.presetsButtons)
        self.addPresetButton.clicked.connect(self.__add_preset)
        self.presetsButtons.layout().addWidget(self.addPresetButton)

        self.renamePresetButton = QPushButton(self.presetsButtons)
        self.renamePresetButton.clicked.connect(self.__rename_preset)
        self.presetsButtons.layout().addWidget(self.renamePresetButton)

        self.editPresetButton = QPushButton(self.presetsButtons)
        self.editPresetButton.clicked.connect(self.__edit_preset)
        self.presetsButtons.layout().addWidget(self.editPresetButton)

        self.removePresetButton = QPushButton(self.presetsButtons)
        self.removePresetButton.clicked.connect(self.__remove_preset)
        self.presetsButtons.layout().addWidget(self.removePresetButton)

        self.cueFromSelectedButton = QPushButton(self.presetsButtons)
        self.cueFromSelectedButton.clicked.connect(self.__cue_from_selected)
        self.presetsButtons.layout().addWidget(self.cueFromSelectedButton)

        self.loadOnSelectedButton = QPushButton(self.presetsButtons)
        self.loadOnSelectedButton.clicked.connect(self.__load_on_selected)
        self.presetsButtons.layout().addWidget(self.loadOnSelectedButton)

        # Import/Export buttons
        self.ieButtons = QWidget(self)
        self.ieButtons.setLayout(QHBoxLayout())
        self.ieButtons.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self.ieButtons, 1, 0)
        self.layout().setAlignment(self.ieButtons, Qt.AlignLeft)

        self.exportSelectedButton = QPushButton(self.ieButtons)
        self.exportSelectedButton.clicked.connect(self.__export_presets)
        self.ieButtons.layout().addWidget(self.exportSelectedButton)

        self.importButton = QPushButton(self.ieButtons)
        self.importButton.clicked.connect(self.__import_presets)
        self.ieButtons.layout().addWidget(self.importButton)

        # Dialog buttons
        self.dialogButtons = QDialogButtonBox(self)
        self.dialogButtons.setStandardButtons(QDialogButtonBox.Ok)
        self.dialogButtons.accepted.connect(self.accept)
        self.layout().addWidget(self.dialogButtons, 1, 1)

        self.layout().setColumnStretch(0, 4)
        self.layout().setColumnStretch(1, 1)

        self.retranslateUi()

        self.__populate()
        self.__selection_changed()

    def retranslateUi(self):
        self.addPresetButton.setText(translate('Presets', 'Add'))
        self.renamePresetButton.setText(translate('Presets', 'Rename'))
        self.editPresetButton.setText(translate('Presets', 'Edit'))
        self.removePresetButton.setText(translate('Presets', 'Remove'))
        self.cueFromSelectedButton.setText(translate('Preset', 'Create Cue'))
        self.loadOnSelectedButton.setText(translate('Preset', 'Load on selected Cues'))
        self.exportSelectedButton.setText(translate('Presets', 'Export selected'))
        self.importButton.setText(translate('Presets', 'Import'))

    def __populate(self):
        self.presetsList.clear()
        try:
            self.presetsList.addItems(scan_presets())
        except OSError as e:
            scan_presets_error(e, parent=self)

    def __remove_preset(self):
        for item in self.presetsList.selectedItems():
            try:
                delete_preset(item.text())
                self.presetsList.takeItem(self.presetsList.currentRow())
            except OSError as e:
                delete_preset_error(e, item.text(), parent=self)

    def __add_preset(self):
        dialog = NewPresetDialog(parent=self)
        if dialog.exec_() == QDialog.Accepted:
            preset_name = dialog.get_name()
            cue_type = dialog.get_type()

            exists = preset_exists(preset_name)
            if not (exists and not check_override_dialog(preset_name)):
                try:
                    write_preset(preset_name, {'_type_': cue_type})
                    if not exists:
                        self.presetsList.addItem(preset_name)
                except OSError as e:
                    write_preset_error(e, preset_name, parent=self)

    def __rename_preset(self):
        item = self.presetsList.currentItem()
        if item is not None:
            new_name = save_preset_dialog(base_name=item.text())
            if new_name is not None and new_name != item.text():
                if preset_exists(new_name):
                    QMessageBox.warning(
                        self,
                        translate('Presets', 'Warning'),
                        translate('Presets', 'The same name is already used!')
                    )
                else:
                    try:
                        rename_preset(item.text(), new_name)
                        item.setText(new_name)
                    except OSError as e:
                        rename_preset_error(e, item.text(), parent=self)

    def __edit_preset(self):
        item = self.presetsList.currentItem()
        if item is not None:
            try:
                preset = load_preset(item.text())
                if preset is not None:
                    try:
                        cue_class = CueFactory.create_cue(preset.get('_type_'))
                        cue_class = cue_class.__class__
                    except Exception:
                        cue_class = Cue

                    edit_dialog = CueSettings(cue_class=cue_class)
                    edit_dialog.load_settings(preset)
                    if edit_dialog.exec_() == edit_dialog.Accepted:
                        preset.update(edit_dialog.get_settings())
                        try:
                            write_preset(item.text(), preset)
                        except OSError as e:
                            write_preset_error(e, item.text(), parent=self)
            except OSError as e:
                load_preset_error(e, item.text(), parent=self)

    def __cue_from_preset(self, preset_name):
        try:
            preset = load_preset(preset_name)
            if preset is not None:
                if CueFactory.has_factory(preset.get('_type_')):
                    cue = CueFactory.create_cue(preset['_type_'])

                    cue.update_properties(preset)
                    Application().cue_model.add(cue)
                else:
                    QMessageBox.warning(
                        self,
                        translate('Presets', 'Warning'),
                        translate('Presets', 'Cannot create a cue from this '
                                             'preset: {}').format(preset_name)
                    )
        except OSError as e:
            load_preset_error(e, preset_name, parent=self)

    def __cue_from_selected(self):
        for item in self.presetsList.selectedItems():
            self.__cue_from_preset(item.text())

    def __load_on_selected(self):
        item = self.presetsList.currentItem()
        if item is not None:
            preset_name = item.text()
            try:
                cues = Application().layout.get_selected_cues()
                if cues:
                    load_on_cues(preset_name, cues)
            except OSError as e:
                load_preset_error(e, preset_name, parent=MainWindow())

    def __export_presets(self):
        names = [item.text() for item in self.presetsList.selectedItems()]
        archive, _ = QFileDialog.getSaveFileName(
            self,
            directory='archive.presets',
            filter='*.presets'
        )

        if archive != '':
            if not archive.endswith('.presets'):
                archive += '.presets'
            try:
                export_presets(names, archive)
            except PresetExportError as e:
                QDetailedMessageBox.dcritical(
                    translate('Presets', 'Presets'),
                    translate('Presets', 'Cannot export correctly.'),
                    str(e),
                    parent=self
                )

    def __import_presets(self):
        archive, _ = QFileDialog.getOpenFileName(self, filter='*.presets')
        if archive != '':
            try:
                if import_has_conflicts(archive):
                    answer = QMessageBox.question(
                        self,
                        translate('Presets', 'Presets'),
                        translate('Presets', 'Some presets already exists, '
                                             'overwrite?'),
                        buttons=QMessageBox.Yes | QMessageBox.Cancel)

                    if answer != QMessageBox.Yes:
                        return

                import_presets(archive)
                self.__populate()
            except PresetImportError as e:
                QDetailedMessageBox.dcritical(
                    translate('Presets', 'Presets'),
                    translate('Presets', 'Cannot import correctly.'),
                    str(e),
                    parent=self
                )

    def __selection_changed(self):
        selection = len(self.presetsList.selectedIndexes())

        self.editPresetButton.setEnabled(selection == 1)
        self.renamePresetButton.setEnabled(selection == 1)
        self.loadOnSelectedButton.setEnabled(selection == 1)
        self.removePresetButton.setEnabled(selection > 0)
        self.cueFromSelectedButton.setEnabled(selection > 0)
        self.exportSelectedButton.setEnabled(selection > 0)


class NewPresetDialog(QDialog):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.resize(320, 105)
        self.setMaximumSize(self.size())
        self.setMinimumSize(self.size())
        self.setModal(True)
        self.setLayout(QGridLayout())

        self.nameLabel = QLabel(self)
        self.layout().addWidget(self.nameLabel, 0, 0)

        self.nameLineEdit = QLineEdit(self)
        self.layout().addWidget(self.nameLineEdit, 0, 1)

        self.typeLabel = QLabel(self)
        self.layout().addWidget(self.typeLabel, 1, 0)

        self.typeComboBox = QComboBox(self)
        for cue_class in CueSettingsRegistry().ref_classes():
            self.typeComboBox.addItem(translate('CueName', cue_class.Name),
                                      cue_class.__name__)
        self.layout().addWidget(self.typeComboBox, 1, 1)

        self.dialogButtons = QDialogButtonBox(self)
        self.dialogButtons.setStandardButtons(QDialogButtonBox.Ok |
                                              QDialogButtonBox.Cancel)
        self.dialogButtons.accepted.connect(self.accept)
        self.dialogButtons.rejected.connect(self.reject)
        self.layout().addWidget(self.dialogButtons, 2, 0, 1, 2)

        self.layout().setColumnStretch(0, 1)
        self.layout().setColumnStretch(1, 3)

        self.retranslateUi()

    def retranslateUi(self):
        self.nameLabel.setText(translate('Presets', 'Preset name'))
        self.typeLabel.setText(translate('Presets', 'Cue type'))

    def get_name(self):
        return self.nameLineEdit.text()

    def get_type(self):
        return self.typeComboBox.currentData()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy


class CueFactory:
    """Provide a generic factory to build different cues types.

    Cues can be register via `register_factory` function.
    """

    __REGISTRY = {}

    # Register methods

    @classmethod
    def register_factory(cls, cue_type, factory):
        """Register a new cue-type in the factory.

        :param cue_type: The cue class name
        :type cue_type: str
        :param factory: The cue class or a factory function
        """
        cls.__REGISTRY[cue_type] = factory

    @classmethod
    def has_factory(cls, cue_type):
        """Return True if there is a factory for `cue_type`

        :param cue_type: The cue type to check
        :rtype cue_type: str
        :rtype: bool
        """
        return cue_type in cls.__REGISTRY

    @classmethod
    def remove_factory(cls, cue_type):
        """Remove the registered cue from the factory

        :param cue_type: the cue class name (the same used for registration)
        """
        cls.__REGISTRY.pop(cue_type)

    # Create methods

    @classmethod
    def create_cue(cls, cue_type, cue_id=None, **kwargs):
        """Return a new cue of the specified type.

        ..note:
            Some factory can take keyword-arguments (e.g. URIAudio)

        :param cue_id: The id to use with the new cue
        :param cue_type: The cue type
        :rtype: lisp.cues.cue.Cue
        """
        factory = cls.__REGISTRY.get(cue_type)

        if not callable(factory):
            raise Exception(
                'Cue not available or badly registered: {}'.format(cue_type))

        return factory(id=cue_id, **kwargs)

    @classmethod
    def clone_cue(cls, cue):
        """Return a copy of the given cue. The id is not copied.

        :param cue: the cue to be copied
        :rtype: lisp.cues.cue.Cue
        """
        properties = deepcopy(cue.properties())
        properties.pop('id')

        cue = cls.create_cue(cue.__class__.__name__)
        cue.update_properties(properties)

        return cue



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.
from itertools import chain

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication

from lisp.core.decorators import memoize


@memoize
def load_icon(icon_name):
    """Return a QIcon from the icon theme.

    The loaded icons are cached.

    .. warning:
        QIcons should be loaded only from the QT main loop.
    """
    return QIcon.fromTheme(icon_name)


@memoize
def pixmap_from_icon(icon_name, size):
    """Return a QPixmap of "size x size" pixels from the icon theme.

    The returned pixmaps are cached.

    .. warning:
        QPixmap should be created only from the QT main loop.
    """
    return load_icon(icon_name).pixmap(size, size)


def qfile_filters(extensions, allexts=True, anyfile=True):
    """Create a filter-string for a FileChooser.

    The result will be something like this: '<group1> (*.ext1 *.ext2);;
    <group2> (*.ext1)'

    :param extensions: The extensions as a dictionary {group: [extensions]}
    :type extensions: dict
    :param allexts: Add a group composed by all the given groups
    :type allexts: bool
    :param anyfile: Add the "Any File" group
    :type anyfile: bool
    :return: A QFileDialog filter-string
    :rtype: str
    """
    filters = []

    for key in extensions:
        filters.append(key.title() + ' (' + ' *.'.join(extensions[key]) + ')')

    filters.sort()

    if allexts:
        filters.insert(0, 'All supported (' +
                       ' *.'.join(chain(*extensions.values())) + ')')
    if anyfile:
        filters.append('Any file (*)')

    return ';;'.join(filters)


def translate(context, text, disambiguation=None, n=-1):
    return QApplication.translate(context, text, disambiguation, n)


def translate_many(context, texts):
    """Return a translate iterator."""
    for item in texts:
        yield translate(context, item)


def tr_sorted(context, iterable, key=None, reverse=False):
    """Return a new sorted list from the items in iterable.

    The sorting is done using translated versions of the iterable values.
    """
    if key is not None:
        def tr_key(item):
            translate(context, key(item))
    else:
        def tr_key(item):
            translate(context, item)

    return sorted(iterable, key=tr_key, reverse=reverse)\
